#!/bin/env tdaq_python

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from queue import Queue
from time import time
import subprocess
import threading
import sys
import os
import datetime
import signal
import importlib

#Imports of script
from cs.Threads.ManagerThread import ManagerThread
from cs.Threads.CopyThread import CopyThread
from cs.Threads.DeleteThread import DeleteThread
from cs.Threads.ERSThread import ERSThread
from cs.Threads.DdmMonitoringThread import DdmMonitoringThread
from cs.Threads.InfoServiceThread import InfoServiceThread
import cs.Tools.ConfigParser as ConfigParser
import cs.Tools.Constants as Constants
import cs.Tools.LogConfig as LogConfig

##### Global flag for exit #####
exitFlag = 0

##### Get Configuration #####
if len(sys.argv) >= 2:
    try:
        config = ConfigParser.ConfigHolder(sys.argv[1])
    except Exception as ex:
        print('Configuration file parsing failed because:', ex)
        sys.exit(1)
else:
    print('You should provide a configuration file! Exiting ...')
    print(sys.argv[0], 'conffile.cfg')
    sys.exit(1)

### Change umask to ensure text files are group
### writeable
os.umask(0o002)

##### Set event #####
event = threading.Event()

##### Set lock for accessing the Metadata database #####
dbLock = threading.Lock()

##### Define copy, delete, clear, etc. queues #####
CopyQueue = Queue()
DeleteQueue = Queue()
ClearQueue = Queue()
DdmMonitoringQueue = Queue(config.DdmMonitoringQueueMaxSize)  # created even if never used

##### Set manager, copy and delete threads #####
manager = ManagerThread(config, event, CopyQueue, DeleteQueue, ClearQueue)
copy = CopyThread(config, event, dbLock, CopyQueue, DeleteQueue, ClearQueue,
                  DdmMonitoringQueue)

delete = None
if config.DeleteEnabled:
    delete = DeleteThread(config, event, dbLock, DeleteQueue, ClearQueue)

ddm_publisher = None
if config.DdmMonitoringEnabled:
    ddm_publisher = DdmMonitoringThread(config, event, DdmMonitoringQueue)

if config.IS_enabled:
    info_service_thread = InfoServiceThread(config, event, copy, delete, manager)


def main(conf):

    LogConfig.enable_root_logger()

    if conf.mailList:
        LogConfig.enable_mail_logging(conf)

    # Get main logger, (Will also log to ERS and mail, when enabled)
    logger = LogConfig.enable_file_logging("main", "mainLog.out",
            conf.LogDir, conf.LogLevel)

    ers_thread = None
    if conf.ERSenabled:
        ers_thread = ERSThread(config, event)

    ##### Create db file logger #####
    dblogger = LogConfig.enable_file_logging("Database","DatabaseLog.out",conf.LogDir,conf.LogLevel)

    #### Extend the PYTHON_PATH ####
    pypath = []
    pypath.extend(conf.pythonPathPrepend)
    pypath.extend(sys.path)
    sys.path = pypath

    #### Load filename parser ####
    try:
        mod = importlib.import_module("cs.Tools.FilenameParsers." + conf.Filenameparser)
    except Exception as ex:
        logger.error('Cannot import parser module %s because: "%s". Giving up!',
                     conf.Filenameparser, str(ex))
        return

    if conf.Filenameparser not in dir(mod):
        logger.error('Cannot find an appropriate class in the parser module.'
                     ' The class should be called "%s". Giving up!', conf.Filenameparser)
        return

    # We do not want to trap exceptions from the module
    # initialization function. So, first we check if it exists
    # then we execute it outside the try-except environment
    initfunction = None
    try:
        initfunction = mod.init
    except AttributeError:
        pass

    if initfunction:
        initfunction(conf.FilenameparserConf)

    parser = getattr(mod, conf.Filenameparser)

    #### Check filename parser capabilities ####

    missing_symbols = [x for x in Constants.needed_parser_symbols if x not in dir(parser)]

    if missing_symbols:
        logger.error('Cannot find all the needed functions in the parser module. '
                     'Missing functions are: %s. Giving up!', str(missing_symbols))
        return

    ### Set Fileparser ###

    manager.setFileParser(parser)

    ##### Set update and exit signal handlers #####
    ##### update: SIGUSR1 = 10, exit: SIGUSR2 = 12 #####
    signal.signal(signal.SIGUSR2, Exit)
    signal.signal(signal.SIGINT, Exit)

    ##### Set Oracle Metadata Database connection #####
    db = None
    db = checkDB(db, logger, dblogger, parser, conf)
    lastdbconn = time()
    DBTimeout = conf.DBTimeout
    DBReconnectTimeout = conf.DBReconnectTimeout


    ##### Setup Kerberos if needed #####
    if conf.keytab:

        krb_setcache(conf.krbcache,logger)
        krb_exp = krb_updatetoken(conf.keytab, conf.krbuser, logger)
        if not krb_exp:
            logger.error('Could not initialize kerberos ticket. Giving up!')
            return


    ##### Start the threads #####
    if ers_thread:
        ers_thread.start()
    if conf.IS_enabled:
        info_service_thread.start()
    manager.start()
    copy.start()
    if delete is not None:
        delete.start()
    if conf.DdmMonitoringEnabled:
        ddm_publisher.start()


    ##### Every DBTimeout check if connection to Oracle database is still good #####
    while not exitFlag:

        #Keep always a fresh connection
        if db and (time()-lastdbconn)>DBReconnectTimeout:
            logger.info('Going to refresh Metadata Database connection')
            dbLock.acquire()
            try:
                db.Reconnect()
            except Exception as ex:
                logger.error(str(ex))
                logger.warning('Could not connect to Metadata Database:')
                db = None

            dbLock.release()
            lastdbconn = time()

        logger.debug('Check for connection to Metadata Database')
        db = checkDB(db, logger, dblogger, parser, conf)
        event.wait(DBTimeout)

        # Update KRB token, if needed:
        if conf.keytab:
            now = datetime.datetime.now()
            if not krb_exp:
                 #Previous update failed, we try again
                krb_exp = krb_updatetoken(conf.keytab, conf.krbuser, logger)
            elif now >= krb_exp or (krb_exp-now).seconds < 3600:
                krb_exp = krb_updatetoken(conf.keytab, conf.krbuser, logger)


        #Check worker states
        if not (manager.isAlive() and copy.isAlive() and (delete is None or delete.isAlive())):
            logger.warning('Inconsistent worker states. Manager-->%s Copy-->%s '\
                           'Delete-->%s. Exiting!', manager.isAlive(),
                           copy.isAlive(), delete.isAlive())
            Exit(None, None)

    # end while

    ##### If exit signal, stop threads in clean way #####
    logger.info('Exit signal (12) received: exit in clean way')
    manager.managerExit()
    copy.copyExit()
    if delete:
        delete.deleteExit()

    if ers_thread:
        ers_thread.ersExit()

    if conf.DdmMonitoringEnabled:
        ddm_publisher.DdmMonitoringExit()
    if conf.IS_enabled:
        info_service_thread.info_service_exit()

    ##### Wait for the threads to finish and then stop #####
    manager.join()
    copy.join()
    logger.info('Manager and Copy Threads joined the main thread')
    if delete:
        delete.join()
        logger.info('Delete Thread joined the main thread')
    if ers_thread:
        ers_thread.join()
        logger.info('ERS thread joined the main thread')
    if conf.IS_enabled:
        info_service_thread.join()
        logger.info('InfoService Thread joined the main thread')
    if conf.DdmMonitoringEnabled:
        ddm_publisher.join()
        logger.info('DdmMonitoring Thread joined the main thread')


def Exit(signal_number, stackframe):
    del signal_number, stackframe # unused arguments
    global exitFlag
    exitFlag = 1
    #### If any thread is sleeping, wake it up #####
    event.set()


def checkDB(db, logger, dblogger, parser, conf):
    dbFlag = copy.getDBFlag() and (delete is None or delete.getDBFlag())
    if not dbFlag:
        db = None
    if not db and conf.connection:
        logger.info('Connecting to Metadata Database...')
        try:
            import cs.Tools.Libraries.Database as Database
            db = Database.Database(conf, dblogger, parser)
            logger.info('Connection to Metadata Database succeeded')
        except Exception as ex:
            logger.error(str(ex))
            logger.error('Could not connect to Metadata Database')

            # end try, except
        copy.setDB(db)
        if delete:
            delete.setDB(db)
    # end if not
    return db

    #####  Errors for Oracle Database #####
    #####  StandardError:             #####
    #####     Warning                 #####
    #####     Error:                  #####
    #####       InterfaceError        #####
    #####       DatabaseError:        #####
    #####         DataError           #####
    #####         OperationalError    #####
    #####         IntegrityError      #####
    #####         InternalError       #####
    #####         ProgrammingError    #####
    #####         NotSupportedError   #####


def krb_setcache(krbcachefile, logger):
    try:
        os.environ['KRB5CCNAME'] = krbcachefile
    except TypeError as ex:
        logger.warning('KRB Cache file setting failed: %s ' % str(ex))


def krb_updatetoken(keytab, user, logger):
    kinit = subprocess.Popen(['kinit', '-kt', keytab, user], #'-l','3900s'],
                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    ret = kinit.wait()

    if ret:
        logger.warning('kinit failed. Ret code: %d Output: %s' %
                       (ret, kinit.stdout.read()))
        return None
    else:
        logger.debug('kinit succeeded. Output: %s' % kinit.stdout.read())
        return krb_tokenexpiration(logger)


def krb_tokenexpiration(logger):
    # on SLC6 klist (v1.10.3) does not take LC_ALL into account
    # on CentOS 9 klist (v1.14.1) DOES take it into account
    # This resulted in discrepancies in their output especially for timestamps
    # So it's safer to always set LC_ALL to a 'default' constant
    envm = os.environ.copy()
    envm['LC_ALL'] = 'C'
    klist = subprocess.Popen(['klist',], env = envm, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
    ret = klist.wait()
    out = klist.stdout.read()
    logger.debug('klist done. Return code: %d  Output: %s' % (ret,out))

    if ret:
        logger.critical('klist failed: retcode=%s, output=%s', str(ret), out)
        return datetime.datetime(1970, 1, 1)

    ticket = None
    out = out.split('\n')
    for idx,l in enumerate(out):
        if 'Valid' in l:
            ticket = out[idx+1]
            break

    if ticket is None:
        logger.critical('error parsing klist output: output=%s', out)
        return datetime.datetime(1970, 1, 1)

    ticket = ticket.split()
    logger.debug('Ticket expiration: %s' % ticket)
    try:
        expiration_date = datetime.datetime.strptime(' '.join(ticket[2:4]),
                '%m/%d/%y %H:%M:%S')
    except ValueError as exc:
        logger.critical('error converting string to date: ticket="%s", date_string="%s"',
                ticket,  ' '.join(ticket[2:4]))
        raise exc
    return expiration_date


if __name__ == '__main__':
    main(config)
