#!/bin/env python

from builtins import str
from builtins import object
import cx_Oracle 
import sys
import datetime
import time
import coral_auth


class OraDB(object):

    def __init__(self,connection):
        self.connstring = connection
        self.conn_params = coral_auth.get_connection_parameters_from_connection_string(self.connstring) #pylint: disable=no-member
        self.orcl = cx_Oracle.connect(*self.conn_params) #pylint: disable=no-member
        self.curs = self.orcl.cursor()
        self.lastKeys = {}

    def execute(self,args,keys):
        self.lastKeys = keys
        self.curs.execute(*args,**keys)
        self.orcl.commit()
        return self.curs.rowcount

    def getNextRow(self):
        return self.curs.fetchone()

    def getDateVar(self):
        return self.curs.var(cx_Oracle.DATETIME) #pylint: disable=no-member

    def lastOp(self):
        return self.curs.statement

    def getLastKeys(self):
        return self.lastKeys

    def refresh(self):
        self.orcl.commit()
        self.orcl.close()
        self.orcl = cx_Oracle.connect(*self.conn_params) #pylint: disable=no-member
        self.curs = self.orcl.cursor()

class Database(object):

    def __init__(self,conf,logger,parser):
        self.conf = conf
        self.parser = parser

        ##### Set Database parameters #####
        connection = self.conf.connection
        self.db = OraDB(connection)
        self.file_table = self.conf.file_table
        self.lb_table = self.conf.lb_table
        self.run_table = self.conf.run_table

        #### Set Logger for Database #####
        self.logger = logger

    def Reconnect(self):
        self.db.refresh()
        self.logger.info("DB connection refreshed.")

    def FileInfo(self, sfofile):

        sql="select checksum,filesize,filehealth from " + self.file_table + " "
        sql+="where sfopfn = :ssfopfn"
        args=[]
        keys={}
        keys['ssfopfn']=sfofile
        args.append(sql)
        result = (None, None, None)
        try:
            self.db.execute(args,keys)
            row = self.db.getNextRow()
            if row is None:
                # cannot find the file in the DB
                return (None, None, None)
            result = row[:3]
        except (cx_Oracle.InterfaceError,cx_Oracle.DatabaseError) as ex: #pylint: disable=no-member
            self.logger.error(str(ex))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))
        except:
            ex_info = sys.exc_info()
            self.logger.error(str(ex_info))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))

        return result

    def Deletion(self,sfofile):
        try:
            self.filedeletion(sfofile)
            return True
        except (cx_Oracle.InterfaceError,cx_Oracle.DatabaseError) as ex: #pylint: disable=no-member
            self.logger.error(str(ex))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))
            return False
        except:
            ex_info = sys.exc_info()
            self.logger.error(str(ex_info))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))
            return True

    def Transfer(self, sfofile, castorfile, svcclass):

        retry = True
        while retry:
            retry = False
            try:
                self.filetransfer(sfofile, castorfile, svcclass)
            except (cx_Oracle.InterfaceError, cx_Oracle.DatabaseError) as ex: #pylint: disable=no-member
                ex_info = str(ex)
                self.logger.error(ex_info)
                self.logger.error(str(self.db.lastOp()))
                self.logger.error(str(self.db.getLastKeys()))

                if 'ORA-20199' in ex_info and 'NEXTVAL' in ex_info:
                    retry = True
                else:
                    return False

            except:
                ex_info = sys.exc_info()
                self.logger.error(str(ex_info))
                self.logger.error(str(self.db.lastOp()))
                self.logger.error(str(self.db.getLastKeys()))
                return True


        try:
            parsed=self.parser(sfofile)

            if self.notTransFiles(parsed)==0:
                self.lbtransfer(parsed)

            if self.notTransLBs(parsed)==0:
                self.runtransfer(parsed)

            return True
        except (cx_Oracle.InterfaceError, cx_Oracle.DatabaseError) as ex: #pylint: disable=no-member
            self.logger.error(str(ex))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))
            return False

        except:
            ex_info = sys.exc_info()
            self.logger.error(str(ex_info))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))
            return True

    def UpdateTruncated(self, sfofile, size, checksum):

        sql="update " + self.file_table + " "
        sql+="set filesize=:ssize, checksum=:schecksum "
        sql+="where sfopfn = :ssfopfn"
        args=[]
        keys={}
        keys['ssfopfn']=sfofile
        keys['ssize']=size
        keys['schecksum']=checksum.upper()
        args.append(sql)

        try:
            self.db.execute(args,keys)
            return True
        except (cx_Oracle.InterfaceError,cx_Oracle.DatabaseError) as ex: #pylint: disable=no-member
            self.logger.error(str(ex))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))
        except:
            ex_info = sys.exc_info()
            self.logger.error(str(ex_info))
            self.logger.error(str(self.db.lastOp()))
            self.logger.error(str(self.db.getLastKeys()))

        return False


    def filedeletion(self,sfofile):
        sql="update " + self.file_table + " "
        sql+="set filestate='DELETED', t_stamp = :newdate "
        sql+="where sfopfn = :ssfopfn and filestate = 'CLOSED'"
        args=[]
        keys={}
        keys['newdate']=datetime.datetime.utcfromtimestamp(time.time())
        keys['ssfopfn']=sfofile
        args.insert(0,sql)
        nrow = self.db.execute(args,keys)
        if nrow != 1:
            self.logger.error('Deletion of file: %s. Not exactly one entry updated in file table: %s', sfofile, nrow)
        else:
            self.logger.debug('File table. File state updated to DELETED for file: ' + sfofile)
        # end if,else

    def filetransfer(self, sfofile, castorfile, svcclass):
        sql = "update "+ self.file_table + " "
        sql += "set transferstate='TRANSFERRED', t_stamp = :newdate, "
        sql += "pfn = :newpfn, svcclass = :newsvcclass "
        sql += "where sfopfn = :ssfopfn and filestate = 'CLOSED'"

        args=[]
        keys={}
        keys['newdate'] = datetime.datetime.utcfromtimestamp(time.time())
        keys['newpfn'] = castorfile
        keys['ssfopfn'] = sfofile
        keys['newsvcclass'] = svcclass
        args.insert(0,sql)

        nrow = self.db.execute(args,keys)

        if nrow != 1:
            self.logger.critical('transfer of file %s to %s: not exactly one entry updated in file table (%s)',
                    sfofile, castorfile, str(nrow))
        else:
            self.logger.debug('File table. File state updated to TRANSFERRED for file: ' + sfofile)
        # end if,else

    def lbtransfer(self,parsed):

        keys={}
        keys['ssfo']=parsed.AppId()
        keys['srunnr']=parsed.RunNr()
        keys['slbnr']=parsed.LBNr()
        keys['sstreamt']=parsed.StreamType()
        keys['sstreamn']=parsed.StreamName()

        sql="select state from " + self.lb_table + " "
        sql+="where sfoid = :ssfo and runnr = :srunnr and lumiblocknr = :slbnr"
        sql+=" and streamtype = :sstreamt and stream = :sstreamn"

        args = [sql,]
        self.db.execute(args,keys)
        state = self.db.getNextRow()[0]

        if state == 'CLOSED':
            sql="update "+ self.lb_table + " "
            sql+="set state='TRANSFERRED', t_stamp = :newdate "
            sql+="where sfoid = :ssfo and runnr = :srunnr "
            sql+=" and lumiblocknr = :slbnr"
            sql+=" and streamtype = :sstreamt and stream = :sstreamn "
            sql+=" and state = 'CLOSED'"

            keys['newdate']=datetime.datetime.utcfromtimestamp(time.time())

            args = [sql,]
            nrow = self.db.execute(args,keys)

            if nrow != 1:
                self.logger.error('Not exactly one entry updated in '+
                                  'lumiblock table for SFO '
                                  + str(parsed.AppId())+
                                  ', run ' + str(parsed.RunNr()) +
                                  ', lumiblock ' + str(parsed.LBNr()) +
                                  ', stream type ' + str(parsed.StreamType()) +
                                  ', stream name ' + str(parsed.StreamName()))
            else:
                self.logger.debug('Lumiblock table. Lumiblock state updated '+
                                  'to TRANSFERRED for SFO '
                                  + str(parsed.AppId()) +
                                  ', run ' + str(parsed.RunNr()) +
                                  ', lumiblock ' + str(parsed.LBNr()) +
                                  ', stream type ' + str(parsed.StreamType()) +
                                  ', stream name ' + str(parsed.StreamName()))
            # end if,else

    def runtransfer(self,parsed):

        keys={}
        keys['ssfo']=parsed.AppId()
        keys['srunnr']=parsed.RunNr()
        keys['sstreamt']=parsed.StreamType()
        keys['sstreamn']=parsed.StreamName()

        sql="select state from "+ self.run_table + " "
        sql+="where sfoid = :ssfo and runnr = :srunnr "
        sql+=" and streamtype = :sstreamt and stream = :sstreamn "

        args=[sql,]
        self.db.execute(args,keys)
        state = self.db.getNextRow()[0]

        if state == 'CLOSED':

            sql="update "+ self.run_table + " "
            sql+="set state='TRANSFERRED', t_stamp = :newdate "
            sql+="where sfoid = :ssfo and runnr = :srunnr "
            sql+=" and streamtype = :sstreamt and stream = :sstreamn "
            sql+=" and state = 'CLOSED'"


            keys['newdate']=datetime.datetime.utcfromtimestamp(time.time())

            args = [sql,]
            nrow = self.db.execute(args,keys)
            if nrow != 1:
                self.logger.error('Not exactly one entry updated'
                                  ' in run table '+
                                  'for SFO ' + str(parsed.AppId()) +
                                  ', run ' + str(parsed.RunNr()) +
                                  ', stream type ' + str(parsed.StreamType()) +
                                  ', stream name ' + str(parsed.StreamName()))
            else:
                self.logger.debug('Run table. Run state updated '
                                  + 'to TRANSFERRED ' +
                                  'for SFO ' + str(parsed.AppId()) +
                                  ', run ' + str(parsed.RunNr()) +
                                  ', stream type ' + str(parsed.StreamType()) +
                                  ', stream name ' + str(parsed.StreamName()))
            # end if,else


    def notTransFiles(self,parsed):
        sql="select count(lfn) from " + self.file_table + " "
        sql+="where sfoid = :ssfo and runnr = :srunnr and lumiblocknr = :slbnr "
        sql+=" and streamtype = :sstreamt and stream = :sstreamn "
        sql+=" and transferstate = 'ONDISK'"
        args=[]
        args.insert(0,sql)
        keys={}
        keys['ssfo']=parsed.AppId()
        keys['srunnr']=parsed.RunNr()
        keys['slbnr']=parsed.LBNr()
        keys['sstreamt']=parsed.StreamType()
        keys['sstreamn']=parsed.StreamName()
        self.db.execute(args,keys)
        self.logger.debug('Count number of files ONDISK for SFO ' +
                          str(parsed.AppId()) +
                          ', run ' + str(parsed.RunNr()) +
                          ', lumiblock ' + str(parsed.LBNr()) +
                          ', stream type ' + str(parsed.StreamType()) +
                          ', stream name ' + str(parsed.StreamName()))

        return self.db.getNextRow()[0]

    def notTransLBs(self,parsed):
        sql="select count(state) from " + self.lb_table + " "
        sql+="where sfoid = :ssfo and runnr = :srunnr "
        sql+=" and streamtype = :sstreamt and stream = :sstreamn "
        sql+=" and state <> 'TRANSFERRED'"
        args=[]
        args.insert(0,sql)
        keys={}
        keys['ssfo']=parsed.AppId()
        keys['srunnr']=parsed.RunNr()
        keys['sstreamt']=parsed.StreamType()
        keys['sstreamn']=parsed.StreamName()
        self.db.execute(args,keys)
        self.logger.debug('Count number of NOT TRANSFERRED ' +
                          'lumiblocks for SFO ' + str(parsed.AppId()) +
                          ', run ' + str(parsed.RunNr()) +
                          ', stream type ' + str(parsed.StreamType()) +
                          ', stream name ' + str(parsed.StreamName()))

        return self.db.getNextRow()[0]

    def dbConf(self,conf):
        self.logger.info('Update configuration')
        self.conf = conf
